@@ -3,12 +3,12 @@
  * pg_operator.c
  *	  routines to support manipulation of the pg_operator relation
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/catalog/pg_operator.c,v 1.77 2002/09/04 20:31:14 momjian Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/catalog/pg_operator.c,v 1.84 2003/09/25 06:57:58 petere Exp $
  *
  * NOTES
  *	  these routines moved here from commands/define.c and somewhat cleaned up.
@@ -80,7 +80,7 @@ validOperatorName(const char *name)
 
 	/* Can't contain any invalid characters */
 	/* Test string here should match op_chars in scan.l */
-	if (strspn(name, "~!@#^&|`?$+-*/%<>=") != len)
+	if (strspn(name, "~!@#^&|`?+-*/%<>=") != len)
 		return false;
 
 	/* Can't contain slash-star or dash-dash (comment starts) */
@@ -102,7 +102,7 @@ validOperatorName(const char *name)
 
 		for (ic = len - 2; ic >= 0; ic--)
 		{
-			if (strchr("~!@#^&|`?$%", name[ic]))
+			if (strchr("~!@#^&|`?%", name[ic]))
 				break;
 		}
 		if (ic < 0)
@@ -175,7 +175,7 @@ OperatorLookup(List *operatorName,
 	RegProcedure oprcode;
 
 	operatorObjectId = LookupOperName(operatorName, leftObjectId,
-									  rightObjectId);
+									  rightObjectId, true);
 	if (!OidIsValid(operatorObjectId))
 	{
 		*defined = false;
@@ -212,7 +212,10 @@ OperatorShellMake(const char *operatorName,
 	 * validate operator name
 	 */
 	if (!validOperatorName(operatorName))
-		elog(ERROR, "\"%s\" is not a valid operator name", operatorName);
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_NAME),
+				 errmsg("\"%s\" is not a valid operator name",
+						operatorName)));
 
 	/*
 	 * initialize our *nulls and *values arrays
@@ -398,22 +401,35 @@ OperatorCreate(const char *operatorName,
 	 * Sanity checks
 	 */
 	if (!validOperatorName(operatorName))
-		elog(ERROR, "\"%s\" is not a valid operator name", operatorName);
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_NAME),
+				 errmsg("\"%s\" is not a valid operator name",
+						operatorName)));
 
 	if (!OidIsValid(leftTypeId) && !OidIsValid(rightTypeId))
-		elog(ERROR, "at least one of leftarg or rightarg must be specified");
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_FUNCTION_DEFINITION),
+		errmsg("at least one of leftarg or rightarg must be specified")));
 
 	if (!(OidIsValid(leftTypeId) && OidIsValid(rightTypeId)))
 	{
 		/* If it's not a binary op, these things mustn't be set: */
 		if (commutatorName)
-			elog(ERROR, "only binary operators can have commutators");
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_FUNCTION_DEFINITION),
+				  errmsg("only binary operators can have commutators")));
 		if (joinName)
-			elog(ERROR, "only binary operators can have join selectivity");
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_FUNCTION_DEFINITION),
+			 errmsg("only binary operators can have join selectivity")));
 		if (canHash)
-			elog(ERROR, "only binary operators can hash");
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_FUNCTION_DEFINITION),
+					 errmsg("only binary operators can hash")));
 		if (leftSortName || rightSortName || ltCompareName || gtCompareName)
-			elog(ERROR, "only binary operators can mergejoin");
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_FUNCTION_DEFINITION),
+					 errmsg("only binary operators can merge join")));
 	}
 
 	operatorObjectId = OperatorGet(operatorName,
@@ -423,8 +439,10 @@ OperatorCreate(const char *operatorName,
 								   &operatorAlreadyDefined);
 
 	if (operatorAlreadyDefined)
-		elog(ERROR, "OperatorDef: operator \"%s\" already defined",
-			 operatorName);
+		ereport(ERROR,
+				(errcode(ERRCODE_DUPLICATE_FUNCTION),
+				 errmsg("operator %s already exists",
+						operatorName)));
 
 	/*
 	 * At this point, if operatorObjectId is not InvalidOid then we are
@@ -453,9 +471,7 @@ OperatorCreate(const char *operatorName,
 		typeId[1] = rightTypeId;
 		nargs = 2;
 	}
-	procOid = LookupFuncName(procedureName, nargs, typeId);
-	if (!OidIsValid(procOid))
-		func_error("OperatorDef", procedureName, nargs, typeId, NULL);
+	procOid = LookupFuncName(procedureName, nargs, typeId, false);
 	operResultType = get_func_rettype(procOid);
 
 	/*
@@ -469,9 +485,7 @@ OperatorCreate(const char *operatorName,
 		typeId[2] = INTERNALOID;	/* args list */
 		typeId[3] = INT4OID;	/* varRelid */
 
-		restOid = LookupFuncName(restrictionName, 4, typeId);
-		if (!OidIsValid(restOid))
-			func_error("OperatorDef", restrictionName, 4, typeId, NULL);
+		restOid = LookupFuncName(restrictionName, 4, typeId, false);
 	}
 	else
 		restOid = InvalidOid;
@@ -485,10 +499,9 @@ OperatorCreate(const char *operatorName,
 		typeId[0] = INTERNALOID;	/* Query */
 		typeId[1] = OIDOID;		/* operator OID */
 		typeId[2] = INTERNALOID;	/* args list */
+		typeId[3] = INT2OID;	/* jointype */
 
-		joinOid = LookupFuncName(joinName, 3, typeId);
-		if (!OidIsValid(joinOid))
-			func_error("OperatorDef", joinName, 3, typeId, NULL);
+		joinOid = LookupFuncName(joinName, 4, typeId, false);
 	}
 	else
 		joinOid = InvalidOid;
@@ -620,7 +633,7 @@ OperatorCreate(const char *operatorName,
 								 ObjectIdGetDatum(operatorObjectId),
 								 0, 0, 0);
 		if (!HeapTupleIsValid(tup))
-			elog(ERROR, "OperatorDef: operator %u not found",
+			elog(ERROR, "cache lookup failed for operator %u",
 				 operatorObjectId);
 
 		tup = heap_modifytuple(tup,
@@ -708,7 +721,9 @@ get_other_operator(List *otherOp, Oid otherLeftTypeId, Oid otherRightTypeId,
 		 * only self-linkage for commutation makes sense.
 		 */
 		if (!isCommutator)
-			elog(ERROR, "operator cannot be its own negator or sort operator");
+			ereport(ERROR,
+					(errcode(ERRCODE_INVALID_FUNCTION_DEFINITION),
+					 errmsg("operator cannot be its own negator or sort operator")));
 		return InvalidOid;
 	}
 
@@ -717,16 +732,13 @@ get_other_operator(List *otherOp, Oid otherLeftTypeId, Oid otherRightTypeId,
 	aclresult = pg_namespace_aclcheck(otherNamespace, GetUserId(),
 									  ACL_CREATE);
 	if (aclresult != ACLCHECK_OK)
-		aclcheck_error(aclresult, get_namespace_name(otherNamespace));
+		aclcheck_error(aclresult, ACL_KIND_NAMESPACE,
+					   get_namespace_name(otherNamespace));
 
 	other_oid = OperatorShellMake(otherName,
 								  otherNamespace,
 								  otherLeftTypeId,
 								  otherRightTypeId);
-	if (!OidIsValid(other_oid))
-		elog(ERROR,
-			 "OperatorDef: can't create operator shell \"%s\"",
-			 NameListToString(otherOp));
 	return other_oid;
 }
 
