@@ -3,11 +3,11 @@
  * rewriteHandler.c
  *		Primary module of query rewriter.
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/rewrite/rewriteHandler.c,v 1.113.2.2 2004/01/14 03:39:36 tgl Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/rewrite/rewriteHandler.c,v 1.130.2.1 2004/01/14 03:39:29 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -33,6 +33,13 @@
 #include "utils/lsyscache.h"
 
 
+/* We use a list of these to detect recursion in RewriteQuery */
+typedef struct rewrite_event
+{
+	Oid			relation;		/* OID of relation having rules */
+	CmdType		event;			/* type of rule being fired */
+} rewrite_event;
+
 static Query *rewriteRuleAction(Query *parsetree,
 				  Query *rule_action,
 				  Node *rule_qual,
@@ -41,11 +48,12 @@ static Query *rewriteRuleAction(Query *parsetree,
 static List *adjustJoinTreeList(Query *parsetree, bool removert, int rt_index);
 static void rewriteTargetList(Query *parsetree, Relation target_relation);
 static TargetEntry *process_matched_tle(TargetEntry *src_tle,
-					TargetEntry *prior_tle);
+										TargetEntry *prior_tle,
+										const char *attrName);
 static void markQueryForUpdate(Query *qry, bool skipOldNew);
 static List *matchLocks(CmdType event, RuleLock *rulelocks,
 		   int varno, Query *parsetree);
-static Query *fireRIRrules(Query *parsetree);
+static Query *fireRIRrules(Query *parsetree, List *activeRIRs);
 
 
 /*
@@ -99,9 +107,25 @@ rewriteRuleAction(Query *parsetree,
 	 * Generate expanded rtable consisting of main parsetree's rtable plus
 	 * rule action's rtable; this becomes the complete rtable for the rule
 	 * action.	Some of the entries may be unused after we finish
-	 * rewriting, but if we tried to remove them we'd have a much
-	 * harder job to adjust RT indexes in the query's Vars.  It's OK to
-	 * have unused RT entries, since planner will ignore them.
+	 * rewriting, but we leave them all in place for two reasons:
+	 *
+	 *		* We'd have a much harder job to adjust the query's varnos
+	 *		  if we selectively removed RT entries.
+	 *
+	 *		* If the rule is INSTEAD, then the original query won't be
+	 *		  executed at all, and so its rtable must be preserved so that
+	 *		  the executor will do the correct permissions checks on it.
+	 *
+	 * RT entries that are not referenced in the completed jointree will be
+	 * ignored by the planner, so they do not affect query semantics.  But
+	 * any permissions checks specified in them will be applied during
+	 * executor startup (see ExecCheckRTEPerms()).  This allows us to check
+	 * that the caller has, say, insert-permission on a view, when the view
+	 * is not semantically referenced at all in the resulting query.
+	 *
+	 * When a rule is not INSTEAD, the permissions checks done on its copied
+	 * RT entries will be redundant with those done during execution of the
+	 * original query, but we don't bother to treat that case differently.
 	 *
 	 * NOTE: because planner will destructively alter rtable, we must ensure
 	 * that rule action's rtable is separate and shares no substructure
@@ -125,18 +149,33 @@ rewriteRuleAction(Query *parsetree,
 	 * As above, the action's jointree must not share substructure with the
 	 * main parsetree's.
 	 */
-	if (sub_action->jointree != NULL)
+	if (sub_action->commandType != CMD_UTILITY)
 	{
 		bool		keeporig;
 		List	   *newjointree;
 
+		Assert(sub_action->jointree != NULL);
 		keeporig = (!rangeTableEntry_used((Node *) sub_action->jointree,
 										  rt_index, 0)) &&
 			(rangeTableEntry_used(rule_qual, rt_index, 0) ||
 		  rangeTableEntry_used(parsetree->jointree->quals, rt_index, 0));
 		newjointree = adjustJoinTreeList(parsetree, !keeporig, rt_index);
-		sub_action->jointree->fromlist =
-			nconc(newjointree, sub_action->jointree->fromlist);
+		if (newjointree != NIL)
+		{
+			/*
+			 * If sub_action is a setop, manipulating its jointree will do
+			 * no good at all, because the jointree is dummy.  (Perhaps
+			 * someday we could push the joining and quals down to the
+			 * member statements of the setop?)
+			 */
+			if (sub_action->setOperations != NULL)
+				ereport(ERROR,
+						(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+						 errmsg("conditional UNION/INTERSECT/EXCEPT statements are not implemented")));
+
+			sub_action->jointree->fromlist =
+				nconc(newjointree, sub_action->jointree->fromlist);
+		}
 	}
 
 	/*
@@ -205,9 +244,11 @@ adjustJoinTreeList(Query *parsetree, bool removert, int rt_index)
 		{
 			RangeTblRef *rtr = lfirst(jjt);
 
-			if (IsA(rtr, RangeTblRef) &&rtr->rtindex == rt_index)
+			if (IsA(rtr, RangeTblRef) &&
+				rtr->rtindex == rt_index)
 			{
 				newjointree = lremove(rtr, newjointree);
+				/* foreach is safe because we exit loop after lremove... */
 				break;
 			}
 		}
@@ -225,7 +266,9 @@ adjustJoinTreeList(Query *parsetree, bool removert, int rt_index)
  * attributes that have defaults and are not assigned to in the given tlist.
  * (We do not insert anything for default-less attributes, however.  The
  * planner will later insert NULLs for them, but there's no reason to slow
- * down rewriter processing with extra tlist nodes.)
+ * down rewriter processing with extra tlist nodes.)  Also, for both INSERT
+ * and UPDATE, replace explicit DEFAULT specifications with column default
+ * expressions.
  *
  * 2. Merge multiple entries for the same target attribute, or declare error
  * if we can't.  Presently, multiple entries are only allowed for UPDATE of
@@ -268,8 +311,7 @@ rewriteTargetList(Query *parsetree, Relation target_relation)
 			continue;
 
 		/*
-		 * Look for targetlist entries matching this attr.	We match by
-		 * resno, but the resname should match too.
+		 * Look for targetlist entries matching this attr.
 		 *
 		 * Junk attributes are not candidates to be matched.
 		 */
@@ -280,31 +322,58 @@ rewriteTargetList(Query *parsetree, Relation target_relation)
 
 			if (!resdom->resjunk && resdom->resno == attrno)
 			{
-				Assert(strcmp(resdom->resname,
-							  NameStr(att_tup->attname)) == 0);
-				new_tle = process_matched_tle(old_tle, new_tle);
+				new_tle = process_matched_tle(old_tle, new_tle,
+											  NameStr(att_tup->attname));
 				/* keep scanning to detect multiple assignments to attr */
 			}
 		}
 
-		if (new_tle == NULL && commandType == CMD_INSERT)
+		/*
+		 * Handle the two cases where we need to insert a default
+		 * expression: it's an INSERT and there's no tlist entry for the
+		 * column, or the tlist entry is a DEFAULT placeholder node.
+		 */
+		if ((new_tle == NULL && commandType == CMD_INSERT) ||
+		  (new_tle && new_tle->expr && IsA(new_tle->expr, SetToDefault)))
 		{
-			/*
-			 * Didn't find a matching tlist entry; if it's an INSERT, look
-			 * for a default value, and add a tlist entry computing the
-			 * default if we find one.
-			 */
 			Node	   *new_expr;
 
 			new_expr = build_column_default(target_relation, attrno);
 
+			/*
+			 * If there is no default (ie, default is effectively NULL),
+			 * we can omit the tlist entry in the INSERT case, since the
+			 * planner can insert a NULL for itself, and there's no point
+			 * in spending any more rewriter cycles on the entry.  But in
+			 * the UPDATE case we've got to explicitly set the column to
+			 * NULL.
+			 */
+			if (!new_expr)
+			{
+				if (commandType == CMD_INSERT)
+					new_tle = NULL;
+				else
+				{
+					new_expr = (Node *) makeConst(att_tup->atttypid,
+												  att_tup->attlen,
+												  (Datum) 0,
+												  true, /* isnull */
+												  att_tup->attbyval);
+					/* this is to catch a NOT NULL domain constraint */
+					new_expr = coerce_to_domain(new_expr,
+												InvalidOid,
+												att_tup->atttypid,
+												COERCE_IMPLICIT_CAST);
+				}
+			}
+
 			if (new_expr)
 				new_tle = makeTargetEntry(makeResdom(attrno,
 													 att_tup->atttypid,
 													 att_tup->atttypmod,
 									  pstrdup(NameStr(att_tup->attname)),
 													 false),
-										  new_expr);
+										  (Expr *) new_expr);
 		}
 
 		if (new_tle)
@@ -340,8 +409,7 @@ rewriteTargetList(Query *parsetree, Relation target_relation)
 		{
 			/* Let's just make sure we processed all the non-junk items */
 			if (resdom->resno < 1 || resdom->resno > numattrs)
-				elog(ERROR, "rewriteTargetList: bogus resno %d in targetlist",
-					 resdom->resno);
+				elog(ERROR, "bogus resno %d in targetlist", resdom->resno);
 		}
 	}
 
@@ -353,11 +421,12 @@ rewriteTargetList(Query *parsetree, Relation target_relation)
  * Convert a matched TLE from the original tlist into a correct new TLE.
  *
  * This routine detects and handles multiple assignments to the same target
- * attribute.
+ * attribute.  (The attribute name is needed only for error messages.)
  */
 static TargetEntry *
 process_matched_tle(TargetEntry *src_tle,
-					TargetEntry *prior_tle)
+					TargetEntry *prior_tle,
+					const char *attrName)
 {
 	Resdom	   *resdom = src_tle->resdom;
 	Node	   *priorbottom;
@@ -382,20 +451,24 @@ process_matched_tle(TargetEntry *src_tle,
 		((ArrayRef *) prior_tle->expr)->refassgnexpr == NULL ||
 		((ArrayRef *) src_tle->expr)->refrestype !=
 		((ArrayRef *) prior_tle->expr)->refrestype)
-		elog(ERROR, "Multiple assignments to same attribute \"%s\"",
-			 resdom->resname);
+		ereport(ERROR,
+				(errcode(ERRCODE_SYNTAX_ERROR),
+				 errmsg("multiple assignments to same column \"%s\"",
+						attrName)));
 
 	/*
 	 * Prior TLE could be a nest of ArrayRefs if we do this more than
 	 * once.
 	 */
-	priorbottom = ((ArrayRef *) prior_tle->expr)->refexpr;
+	priorbottom = (Node *) ((ArrayRef *) prior_tle->expr)->refexpr;
 	while (priorbottom != NULL && IsA(priorbottom, ArrayRef) &&
 		   ((ArrayRef *) priorbottom)->refassgnexpr != NULL)
-		priorbottom = ((ArrayRef *) priorbottom)->refexpr;
+		priorbottom = (Node *) ((ArrayRef *) priorbottom)->refexpr;
 	if (!equal(priorbottom, ((ArrayRef *) src_tle->expr)->refexpr))
-		elog(ERROR, "Multiple assignments to same attribute \"%s\"",
-			 resdom->resname);
+		ereport(ERROR,
+				(errcode(ERRCODE_SYNTAX_ERROR),
+				 errmsg("multiple assignments to same column \"%s\"",
+						attrName)));
 
 	/*
 	 * Looks OK to nest 'em.
@@ -404,7 +477,7 @@ process_matched_tle(TargetEntry *src_tle,
 	memcpy(newexpr, src_tle->expr, sizeof(ArrayRef));
 	newexpr->refexpr = prior_tle->expr;
 
-	return makeTargetEntry(resdom, (Node *) newexpr);
+	return makeTargetEntry(resdom, (Expr *) newexpr);
 }
 
 
@@ -466,28 +539,28 @@ build_column_default(Relation rel, int attrno)
 		return NULL;			/* No default anywhere */
 
 	/*
-	 * Make sure the value is coerced to the target column type (might not
-	 * be right type yet if it's not a constant!)  This should match the
-	 * parser's processing of non-defaulted expressions --- see
+	 * Make sure the value is coerced to the target column type; this will
+	 * generally be true already, but there seem to be some corner cases
+	 * involving domain defaults where it might not be true. This should
+	 * match the parser's processing of non-defaulted expressions --- see
 	 * updateTargetListEntry().
 	 */
 	exprtype = exprType(expr);
 
-	expr = coerce_to_target_type(expr, exprtype,
+	expr = coerce_to_target_type(NULL,	/* no UNKNOWN params here */
+								 expr, exprtype,
 								 atttype, atttypmod,
 								 COERCION_ASSIGNMENT,
 								 COERCE_IMPLICIT_CAST);
-	/*
-	 * This really shouldn't fail; should have checked the default's
-	 * type when it was created ...
-	 */
 	if (expr == NULL)
-		elog(ERROR, "Column \"%s\" is of type %s"
-			 " but default expression is of type %s"
-			 "\n\tYou will need to rewrite or cast the expression",
-			 NameStr(att_tup->attname),
-			 format_type_be(atttype),
-			 format_type_be(exprtype));
+		ereport(ERROR,
+				(errcode(ERRCODE_DATATYPE_MISMATCH),
+				 errmsg("column \"%s\" is of type %s"
+						" but default expression is of type %s",
+						NameStr(att_tup->attname),
+						format_type_be(atttype),
+						format_type_be(exprtype)),
+		   errhint("You will need to rewrite or cast the expression.")));
 
 	return expr;
 }
@@ -507,8 +580,8 @@ matchLocks(CmdType event,
 	int			nlocks;
 	int			i;
 
-	Assert(rulelocks != NULL);	/* we get called iff there is some lock */
-	Assert(parsetree != NULL);
+	if (rulelocks == NULL)
+		return NIL;
 
 	if (parsetree->commandType != CMD_SELECT)
 	{
@@ -543,18 +616,19 @@ ApplyRetrieveRule(Query *parsetree,
 				  int rt_index,
 				  bool relation_level,
 				  Relation relation,
-				  bool relIsUsed)
+				  bool relIsUsed,
+				  List *activeRIRs)
 {
 	Query	   *rule_action;
 	RangeTblEntry *rte,
 			   *subrte;
 
 	if (length(rule->actions) != 1)
-		elog(ERROR, "ApplyRetrieveRule: expected just one rule action");
+		elog(ERROR, "expected just one rule action");
 	if (rule->qual != NULL)
-		elog(ERROR, "ApplyRetrieveRule: can't handle qualified ON SELECT rule");
+		elog(ERROR, "cannot handle qualified ON SELECT rule");
 	if (!relation_level)
-		elog(ERROR, "ApplyRetrieveRule: can't handle per-attribute ON SELECT rule");
+		elog(ERROR, "cannot handle per-attribute ON SELECT rule");
 
 	/*
 	 * Make a modifiable copy of the view query, and recursively expand
@@ -562,7 +636,7 @@ ApplyRetrieveRule(Query *parsetree,
 	 */
 	rule_action = copyObject(lfirst(rule->actions));
 
-	rule_action = fireRIRrules(rule_action);
+	rule_action = fireRIRrules(rule_action, activeRIRs);
 
 	/*
 	 * VIEWs are really easy --- just plug the view query in as a
@@ -664,7 +738,7 @@ markQueryForUpdate(Query *qry, bool skipOldNew)
  * the SubLink's subselect link with the possibly-rewritten subquery.
  */
 static bool
-fireRIRonSubLink(Node *node, void *context)
+fireRIRonSubLink(Node *node, List *activeRIRs)
 {
 	if (node == NULL)
 		return false;
@@ -673,7 +747,8 @@ fireRIRonSubLink(Node *node, void *context)
 		SubLink    *sub = (SubLink *) node;
 
 		/* Do what we came for */
-		sub->subselect = (Node *) fireRIRrules((Query *) (sub->subselect));
+		sub->subselect = (Node *) fireRIRrules((Query *) sub->subselect,
+											   activeRIRs);
 		/* Fall through to process lefthand args of SubLink */
 	}
 
@@ -682,7 +757,7 @@ fireRIRonSubLink(Node *node, void *context)
 	 * processed subselects of subselects for us.
 	 */
 	return expression_tree_walker(node, fireRIRonSubLink,
-								  (void *) context);
+								  (void *) activeRIRs);
 }
 
 
@@ -691,7 +766,7 @@ fireRIRonSubLink(Node *node, void *context)
  *	Apply all RIR rules on each rangetable entry in a query
  */
 static Query *
-fireRIRrules(Query *parsetree)
+fireRIRrules(Query *parsetree, List *activeRIRs)
 {
 	int			rt_index;
 
@@ -710,7 +785,6 @@ fireRIRrules(Query *parsetree)
 		LOCKMODE	lockmode;
 		bool		relIsUsed;
 		int			i;
-		List	   *l;
 
 		++rt_index;
 
@@ -723,7 +797,7 @@ fireRIRrules(Query *parsetree)
 		 */
 		if (rte->rtekind == RTE_SUBQUERY)
 		{
-			rte->subquery = fireRIRrules(rte->subquery);
+			rte->subquery = fireRIRrules(rte->subquery, activeRIRs);
 			continue;
 		}
 
@@ -795,18 +869,32 @@ fireRIRrules(Query *parsetree)
 		}
 
 		/*
-		 * Now apply them
+		 * If we found any, apply them --- but first check for recursion!
 		 */
-		foreach(l, locks)
+		if (locks != NIL)
 		{
-			rule = lfirst(l);
+			List	   *newActiveRIRs;
+			List	   *l;
 
-			parsetree = ApplyRetrieveRule(parsetree,
-										  rule,
-										  rt_index,
-										  rule->attrno == -1,
-										  rel,
-										  relIsUsed);
+			if (oidMember(RelationGetRelid(rel), activeRIRs))
+				ereport(ERROR,
+						(errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
+						 errmsg("infinite recursion detected in rules for relation \"%s\"",
+								RelationGetRelationName(rel))));
+			newActiveRIRs = lconso(RelationGetRelid(rel), activeRIRs);
+
+			foreach(l, locks)
+			{
+				rule = lfirst(l);
+
+				parsetree = ApplyRetrieveRule(parsetree,
+											  rule,
+											  rt_index,
+											  rule->attrno == -1,
+											  rel,
+											  relIsUsed,
+											  newActiveRIRs);
+			}
 		}
 
 		heap_close(rel, NoLock);
@@ -817,8 +905,8 @@ fireRIRrules(Query *parsetree)
 	 * in the rtable.
 	 */
 	if (parsetree->hasSubLinks)
-		query_tree_walker(parsetree, fireRIRonSubLink, NULL,
-						  QTW_IGNORE_SUBQUERIES);
+		query_tree_walker(parsetree, fireRIRonSubLink, (void *) activeRIRs,
+						  QTW_IGNORE_RT_SUBQUERIES);
 
 	/*
 	 * If the query was marked having aggregates, check if this is still
@@ -832,7 +920,7 @@ fireRIRrules(Query *parsetree)
 		parsetree->hasAggs = checkExprHasAggs((Node *) parsetree);
 		if (parsetree->hasAggs)
 			if (checkExprHasAggs((Node *) parsetree->jointree))
-				elog(ERROR, "fireRIRrules: failed to remove aggs from qual");
+				elog(ERROR, "failed to remove aggregates from qual");
 	}
 	if (parsetree->hasSubLinks)
 		parsetree->hasSubLinks = checkExprHasSubLink((Node *) parsetree);
@@ -903,7 +991,7 @@ CopyAndAddInvertedQual(Query *parsetree,
  * rows that the qualified action doesn't act on.  (If there are multiple
  * qualified INSTEAD rules, we AND all the negated quals onto a single
  * modified original query.)  We won't execute the original, unmodified
- * query if we find either qualified or unqualified INSTEAD rules.  If
+ * query if we find either qualified or unqualified INSTEAD rules.	If
  * we find both, the modified original query is discarded too.
  */
 static List *
@@ -922,7 +1010,7 @@ fireRules(Query *parsetree,
 		RewriteRule *rule_lock = (RewriteRule *) lfirst(i);
 		Node	   *event_qual = rule_lock->qual;
 		List	   *actions = rule_lock->actions;
-		QuerySource	qsrc;
+		QuerySource qsrc;
 		List	   *r;
 
 		/* Determine correct QuerySource value for actions */
@@ -933,7 +1021,7 @@ fireRules(Query *parsetree,
 			else
 			{
 				qsrc = QSRC_INSTEAD_RULE;
-				*instead_flag = true; /* report unqualified INSTEAD */
+				*instead_flag = true;	/* report unqualified INSTEAD */
 			}
 		}
 		else
@@ -947,14 +1035,13 @@ fireRules(Query *parsetree,
 			 * qualifications of the INSTEAD rules are added so it does
 			 * its actions only in cases where the rule quals of all
 			 * INSTEAD rules are false. Think of it as the default action
-			 * in a case. We save this in *qual_product so
-			 * deepRewriteQuery() can add it to the query list after we
-			 * mangled it up enough.
+			 * in a case. We save this in *qual_product so RewriteQuery()
+			 * can add it to the query list after we mangled it up enough.
 			 *
-			 * If we have already found an unqualified INSTEAD rule,
-			 * then *qual_product won't be used, so don't bother building it.
+			 * If we have already found an unqualified INSTEAD rule, then
+			 * *qual_product won't be used, so don't bother building it.
 			 */
-			if (! *instead_flag)
+			if (!*instead_flag)
 			{
 				if (*qual_product == NULL)
 					*qual_product = parsetree;
@@ -977,6 +1064,7 @@ fireRules(Query *parsetree,
 											event_qual, rt_index, event);
 
 			rule_action->querySource = qsrc;
+			rule_action->canSetTag = false;		/* might change later */
 
 			results = lappend(results, rule_action);
 		}
@@ -987,124 +1075,113 @@ fireRules(Query *parsetree,
 
 
 /*
- * One pass of rewriting a single query.
- *
- * parsetree is the input query.  Return value and output parameters
- * are defined the same as for fireRules, above.
- */
-static List *
-RewriteQuery(Query *parsetree, bool *instead_flag, Query **qual_product)
-{
-	CmdType		event;
-	List	   *product_queries = NIL;
-	int			result_relation;
-	RangeTblEntry *rt_entry;
-	Relation	rt_entry_relation;
-	RuleLock   *rt_entry_locks;
-
-	Assert(parsetree != NULL);
-
-	event = parsetree->commandType;
-
-	/*
-	 * SELECT rules are handled later when we have all the queries that
-	 * should get executed
-	 */
-	if (event == CMD_SELECT)
-		return NIL;
-
-	/*
-	 * Utilities aren't rewritten at all - why is this here?
-	 */
-	if (event == CMD_UTILITY)
-		return NIL;
-
-	/*
-	 * the statement is an update, insert or delete - fire rules on it.
-	 */
-	result_relation = parsetree->resultRelation;
-	Assert(result_relation != 0);
-	rt_entry = rt_fetch(result_relation, parsetree->rtable);
-	Assert(rt_entry->rtekind == RTE_RELATION);
-
-	/*
-	 * This may well be the first access to the result relation during the
-	 * current statement (it will be, if this Query was extracted from a
-	 * rule or somehow got here other than via the parser). Therefore,
-	 * grab the appropriate lock type for a result relation, and do not
-	 * release it until end of transaction.  This protects the rewriter
-	 * and planner against schema changes mid-query.
-	 */
-	rt_entry_relation = heap_open(rt_entry->relid, RowExclusiveLock);
-
-	/*
-	 * If it's an INSERT or UPDATE, rewrite the targetlist into standard
-	 * form.  This will be needed by the planner anyway, and doing it now
-	 * ensures that any references to NEW.field will behave sanely.
-	 */
-	if (event == CMD_INSERT || event == CMD_UPDATE)
-		rewriteTargetList(parsetree, rt_entry_relation);
-
-	/*
-	 * Collect and apply the appropriate rules.
-	 */
-	rt_entry_locks = rt_entry_relation->rd_rules;
-
-	if (rt_entry_locks != NULL)
-	{
-		List	   *locks = matchLocks(event, rt_entry_locks,
-									   result_relation, parsetree);
-
-		product_queries = fireRules(parsetree,
-									result_relation,
-									event,
-									locks,
-									instead_flag,
-									qual_product);
-	}
-
-	heap_close(rt_entry_relation, NoLock);		/* keep lock! */
-
-	return product_queries;
-}
-
-
-/*
- * to avoid infinite recursion, we restrict the number of times a query
- * can be rewritten. Detecting cycles is left for the reader as an exercise.
- */
-#ifndef REWRITE_INVOKE_MAX
-#define REWRITE_INVOKE_MAX		100
-#endif
-
-static int	numQueryRewriteInvoked = 0;
-
-/*
- * deepRewriteQuery -
+ * RewriteQuery -
  *	  rewrites the query and apply the rules again on the queries rewritten
+ *
+ * rewrite_events is a list of open query-rewrite actions, so we can detect
+ * infinite recursion.
  */
 static List *
-deepRewriteQuery(Query *parsetree)
+RewriteQuery(Query *parsetree, List *rewrite_events)
 {
-	List	   *rewritten = NIL;
-	List	   *result;
+	CmdType		event = parsetree->commandType;
 	bool		instead = false;
 	Query	   *qual_product = NULL;
-	List	   *n;
+	List	   *rewritten = NIL;
 
-	if (++numQueryRewriteInvoked > REWRITE_INVOKE_MAX)
-		elog(ERROR, "query rewritten %d times, may contain cycles",
-			 numQueryRewriteInvoked - 1);
-
-	result = RewriteQuery(parsetree, &instead, &qual_product);
-
-	foreach(n, result)
+	/*
+	 * If the statement is an update, insert or delete - fire rules on it.
+	 *
+	 * SELECT rules are handled later when we have all the queries that
+	 * should get executed.  Also, utilities aren't rewritten at all (do
+	 * we still need that check?)
+	 */
+	if (event != CMD_SELECT && event != CMD_UTILITY)
 	{
-		Query	   *pt = lfirst(n);
-		List	   *newstuff;
+		int			result_relation;
+		RangeTblEntry *rt_entry;
+		Relation	rt_entry_relation;
+		List	   *locks;
 
-		newstuff = deepRewriteQuery(pt);
-		rewritten = nconc(rewritten, newstuff);
+		result_relation = parsetree->resultRelation;
+		Assert(result_relation != 0);
+		rt_entry = rt_fetch(result_relation, parsetree->rtable);
+		Assert(rt_entry->rtekind == RTE_RELATION);
+
+		/*
+		 * This may well be the first access to the result relation during
+		 * the current statement (it will be, if this Query was extracted
+		 * from a rule or somehow got here other than via the parser).
+		 * Therefore, grab the appropriate lock type for a result
+		 * relation, and do not release it until end of transaction.  This
+		 * protects the rewriter and planner against schema changes
+		 * mid-query.
+		 */
+		rt_entry_relation = heap_open(rt_entry->relid, RowExclusiveLock);
+
+		/*
+		 * If it's an INSERT or UPDATE, rewrite the targetlist into
+		 * standard form.  This will be needed by the planner anyway, and
+		 * doing it now ensures that any references to NEW.field will
+		 * behave sanely.
+		 */
+		if (event == CMD_INSERT || event == CMD_UPDATE)
+			rewriteTargetList(parsetree, rt_entry_relation);
+
+		/*
+		 * Collect and apply the appropriate rules.
+		 */
+		locks = matchLocks(event, rt_entry_relation->rd_rules,
+						   result_relation, parsetree);
+
+		if (locks != NIL)
+		{
+			List	   *product_queries;
+
+			product_queries = fireRules(parsetree,
+										result_relation,
+										event,
+										locks,
+										&instead,
+										&qual_product);
+
+			/*
+			 * If we got any product queries, recursively rewrite them ---
+			 * but first check for recursion!
+			 */
+			if (product_queries != NIL)
+			{
+				List	   *n;
+				rewrite_event *rev;
+
+				foreach(n, rewrite_events)
+				{
+					rev = (rewrite_event *) lfirst(n);
+					if (rev->relation == RelationGetRelid(rt_entry_relation) &&
+						rev->event == event)
+						ereport(ERROR,
+							 (errcode(ERRCODE_INVALID_OBJECT_DEFINITION),
+							  errmsg("infinite recursion detected in rules for relation \"%s\"",
+						   RelationGetRelationName(rt_entry_relation))));
+				}
+
+				rev = (rewrite_event *) palloc(sizeof(rewrite_event));
+				rev->relation = RelationGetRelid(rt_entry_relation);
+				rev->event = event;
+				rewrite_events = lcons(rev, rewrite_events);
+
+				foreach(n, product_queries)
+				{
+					Query	   *pt = (Query *) lfirst(n);
+					List	   *newstuff;
+
+					newstuff = RewriteQuery(pt, rewrite_events);
+					rewritten = nconc(rewritten, newstuff);
+				}
+			}
+		}
+
+		heap_close(rt_entry_relation, NoLock);	/* keep lock! */
 	}
 
 	/*
@@ -1116,9 +1193,9 @@ deepRewriteQuery(Query *parsetree)
 	 * disappear so the scans for them in the rule actions cannot find
 	 * them.
 	 *
-	 * If we found any unqualified INSTEAD, the original query is not
-	 * done at all, in any form.  Otherwise, we add the modified form
-	 * if qualified INSTEADs were found, else the unmodified form.
+	 * If we found any unqualified INSTEAD, the original query is not done at
+	 * all, in any form.  Otherwise, we add the modified form if qualified
+	 * INSTEADs were found, else the unmodified form.
 	 */
 	if (!instead)
 	{
@@ -1142,22 +1219,6 @@ deepRewriteQuery(Query *parsetree)
 }
 
 
-/*
- * QueryRewriteOne -
- *	  rewrite one query
- */
-static List *
-QueryRewriteOne(Query *parsetree)
-{
-	numQueryRewriteInvoked = 0;
-
-	/*
-	 * take a deep breath and apply all the rewrite rules - ay
-	 */
-	return deepRewriteQuery(parsetree);
-}
-
-
 /*
  * QueryRewrite -
  *	  Primary entry point to the query rewriter.
@@ -1173,13 +1234,16 @@ QueryRewrite(Query *parsetree)
 	List	   *querylist;
 	List	   *results = NIL;
 	List	   *l;
+	CmdType		origCmdType;
+	bool		foundOriginalQuery;
+	Query	   *lastInstead;
 
 	/*
 	 * Step 1
 	 *
 	 * Apply all non-SELECT rules possibly getting 0 or many queries
 	 */
-	querylist = QueryRewriteOne(parsetree);
+	querylist = RewriteQuery(parsetree, NIL);
 
 	/*
 	 * Step 2
@@ -1190,7 +1254,7 @@ QueryRewrite(Query *parsetree)
 	{
 		Query	   *query = (Query *) lfirst(l);
 
-		query = fireRIRrules(query);
+		query = fireRIRrules(query, NIL);
 
 		/*
 		 * If the query target was rewritten as a view, complain.
@@ -1205,19 +1269,25 @@ QueryRewrite(Query *parsetree)
 				switch (query->commandType)
 				{
 					case CMD_INSERT:
-						elog(ERROR, "Cannot insert into a view"
-							 "\n\tYou need an unconditional ON INSERT DO INSTEAD rule");
+						ereport(ERROR,
+								(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+								 errmsg("cannot insert into a view"),
+								 errhint("You need an unconditional ON INSERT DO INSTEAD rule.")));
 						break;
 					case CMD_UPDATE:
-						elog(ERROR, "Cannot update a view"
-							 "\n\tYou need an unconditional ON UPDATE DO INSTEAD rule");
+						ereport(ERROR,
+								(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+								 errmsg("cannot update a view"),
+								 errhint("You need an unconditional ON UPDATE DO INSTEAD rule.")));
 						break;
 					case CMD_DELETE:
-						elog(ERROR, "Cannot delete from a view"
-							 "\n\tYou need an unconditional ON DELETE DO INSTEAD rule");
+						ereport(ERROR,
+								(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+								 errmsg("cannot delete from a view"),
+								 errhint("You need an unconditional ON DELETE DO INSTEAD rule.")));
 						break;
 					default:
-						elog(ERROR, "QueryRewrite: unexpected commandType %d",
+						elog(ERROR, "unrecognized commandType: %d",
 							 (int) query->commandType);
 						break;
 				}
@@ -1227,5 +1297,52 @@ QueryRewrite(Query *parsetree)
 		results = lappend(results, query);
 	}
 
+	/*
+	 * Step 3
+	 *
+	 * Determine which, if any, of the resulting queries is supposed to set
+	 * the command-result tag; and update the canSetTag fields
+	 * accordingly.
+	 *
+	 * If the original query is still in the list, it sets the command tag.
+	 * Otherwise, the last INSTEAD query of the same kind as the original
+	 * is allowed to set the tag.  (Note these rules can leave us with no
+	 * query setting the tag.  The tcop code has to cope with this by
+	 * setting up a default tag based on the original un-rewritten query.)
+	 *
+	 * The Asserts verify that at most one query in the result list is marked
+	 * canSetTag.  If we aren't checking asserts, we can fall out of the
+	 * loop as soon as we find the original query.
+	 */
+	origCmdType = parsetree->commandType;
+	foundOriginalQuery = false;
+	lastInstead = NULL;
+
+	foreach(l, results)
+	{
+		Query	   *query = (Query *) lfirst(l);
+
+		if (query->querySource == QSRC_ORIGINAL)
+		{
+			Assert(query->canSetTag);
+			Assert(!foundOriginalQuery);
+			foundOriginalQuery = true;
+#ifndef USE_ASSERT_CHECKING
+			break;
+#endif
+		}
+		else
+		{
+			Assert(!query->canSetTag);
+			if (query->commandType == origCmdType &&
+				(query->querySource == QSRC_INSTEAD_RULE ||
+				 query->querySource == QSRC_QUAL_INSTEAD_RULE))
+				lastInstead = query;
+		}
+	}
+
+	if (!foundOriginalQuery && lastInstead != NULL)
+		lastInstead->canSetTag = true;
+
 	return results;
 }
