@@ -3,12 +3,12 @@
  * md.c
  *	  This code manages relations that reside on magnetic disk.
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/storage/smgr/md.c,v 1.92 2002/08/06 02:36:34 tgl Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/storage/smgr/md.c,v 1.98 2003/08/04 02:40:04 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -249,7 +249,7 @@ mdextend(Relation reln, BlockNumber blocknum, char *buffer)
 	seekpos = (long) (BLCKSZ * (blocknum % ((BlockNumber) RELSEG_SIZE)));
 #ifdef DIAGNOSTIC
 	if (seekpos >= BLCKSZ * RELSEG_SIZE)
-		elog(FATAL, "seekpos too big!");
+		elog(FATAL, "seekpos too big");
 #endif
 #else
 	seekpos = (long) (BLCKSZ * (blocknum));
@@ -284,7 +284,7 @@ mdextend(Relation reln, BlockNumber blocknum, char *buffer)
 #ifndef LET_OS_MANAGE_FILESIZE
 #ifdef DIAGNOSTIC
 	if (_mdnblocks(v->mdfd_vfd, BLCKSZ) > ((BlockNumber) RELSEG_SIZE))
-		elog(FATAL, "segment too big!");
+		elog(FATAL, "segment too big");
 #endif
 #endif
 
@@ -338,7 +338,7 @@ mdopen(Relation reln)
 
 #ifdef DIAGNOSTIC
 	if (_mdnblocks(fd, BLCKSZ) > ((BlockNumber) RELSEG_SIZE))
-		elog(FATAL, "segment too big on relopen!");
+		elog(FATAL, "segment too big");
 #endif
 #endif
 
@@ -421,7 +421,7 @@ mdread(Relation reln, BlockNumber blocknum, char *buffer)
 
 #ifdef DIAGNOSTIC
 	if (seekpos >= BLCKSZ * RELSEG_SIZE)
-		elog(FATAL, "seekpos too big!");
+		elog(FATAL, "seekpos too big");
 #endif
 #else
 	seekpos = (long) (BLCKSZ * (blocknum));
@@ -434,8 +434,12 @@ mdread(Relation reln, BlockNumber blocknum, char *buffer)
 	if ((nbytes = FileRead(v->mdfd_vfd, buffer, BLCKSZ)) != BLCKSZ)
 	{
 		/*
-		 * If we are at EOF, return zeroes without complaining. (XXX Is
-		 * this still necessary/a good idea??)
+		 * If we are at or past EOF, return zeroes without complaining.
+		 * Also substitute zeroes if we found a partial block at EOF.
+		 *
+		 * XXX this is really ugly, bad design.  However the current
+		 * implementation of hash indexes requires it, because hash index
+		 * pages are initialized out-of-order.
 		 */
 		if (nbytes == 0 ||
 			(nbytes > 0 && mdnblocks(reln) == blocknum))
@@ -464,7 +468,7 @@ mdwrite(Relation reln, BlockNumber blocknum, char *buffer)
 	seekpos = (long) (BLCKSZ * (blocknum % ((BlockNumber) RELSEG_SIZE)));
 #ifdef DIAGNOSTIC
 	if (seekpos >= BLCKSZ * RELSEG_SIZE)
-		elog(FATAL, "seekpos too big!");
+		elog(FATAL, "seekpos too big");
 #endif
 #else
 	seekpos = (long) (BLCKSZ * (blocknum));
@@ -503,7 +507,7 @@ mdblindwrt(RelFileNode rnode,
 	seekpos = (long) (BLCKSZ * (blkno % ((BlockNumber) RELSEG_SIZE)));
 #ifdef DIAGNOSTIC
 	if (seekpos >= BLCKSZ * RELSEG_SIZE)
-		elog(FATAL, "seekpos too big!");
+		elog(FATAL, "seekpos too big");
 #endif
 #else
 	seekpos = (long) (BLCKSZ * (blkno));
@@ -512,7 +516,7 @@ mdblindwrt(RelFileNode rnode,
 	errno = 0;
 	if (lseek(fd, seekpos, SEEK_SET) != seekpos)
 	{
-		elog(LOG, "mdblindwrt: lseek(%ld) failed: %m", seekpos);
+		elog(LOG, "lseek(%ld) failed: %m", seekpos);
 		close(fd);
 		return SM_FAIL;
 	}
@@ -526,13 +530,13 @@ mdblindwrt(RelFileNode rnode,
 		/* if write didn't set errno, assume problem is no disk space */
 		if (errno == 0)
 			errno = ENOSPC;
-		elog(LOG, "mdblindwrt: write() failed: %m");
+		elog(LOG, "write() failed: %m");
 		status = SM_FAIL;
 	}
 
 	if (close(fd) < 0)
 	{
-		elog(LOG, "mdblindwrt: close() failed: %m");
+		elog(LOG, "close() failed: %m");
 		status = SM_FAIL;
 	}
 
@@ -547,7 +551,7 @@ mdblindwrt(RelFileNode rnode,
  *		called, then only segments up to the last one actually touched
  *		are present in the chain...
  *
- *		Returns # of blocks, elog's on error.
+ *		Returns # of blocks, ereport's on error.
  */
 BlockNumber
 mdnblocks(Relation reln)
@@ -565,11 +569,27 @@ mdnblocks(Relation reln)
 
 #ifndef LET_OS_MANAGE_FILESIZE
 	segno = 0;
+
+	/*
+	 * Skip through any segments that aren't the last one, to avoid
+	 * redundant seeks on them.  We have previously verified that these
+	 * segments are exactly RELSEG_SIZE long, and it's useless to recheck
+	 * that each time. (NOTE: this assumption could only be wrong if
+	 * another backend has truncated the relation.	We rely on higher code
+	 * levels to handle that scenario by closing and re-opening the md
+	 * fd.)
+	 */
+	while (v->mdfd_chain != (MdfdVec *) NULL)
+	{
+		segno++;
+		v = v->mdfd_chain;
+	}
+
 	for (;;)
 	{
 		nblocks = _mdnblocks(v->mdfd_vfd, BLCKSZ);
 		if (nblocks > ((BlockNumber) RELSEG_SIZE))
-			elog(FATAL, "segment too big in mdnblocks!");
+			elog(FATAL, "segment too big");
 		if (nblocks < ((BlockNumber) RELSEG_SIZE))
 			return (segno * ((BlockNumber) RELSEG_SIZE)) + nblocks;
 
@@ -589,7 +609,7 @@ mdnblocks(Relation reln)
 			 */
 			v->mdfd_chain = _mdfd_openseg(reln, segno, O_CREAT);
 			if (v->mdfd_chain == (MdfdVec *) NULL)
-				elog(ERROR, "cannot count blocks for %s -- open failed: %m",
+				elog(ERROR, "could not count blocks of \"%s\": %m",
 					 RelationGetRelationName(reln));
 		}
 
@@ -836,7 +856,7 @@ _mdfd_openseg(Relation reln, BlockNumber segno, int oflags)
 
 #ifdef DIAGNOSTIC
 	if (_mdnblocks(fd, BLCKSZ) > ((BlockNumber) RELSEG_SIZE))
-		elog(FATAL, "segment too big on openseg!");
+		elog(FATAL, "segment too big");
 #endif
 #endif
 
@@ -855,7 +875,7 @@ _mdfd_getrelnfd(Relation reln)
 	if (fd < 0)
 	{
 		if ((fd = mdopen(reln)) < 0)
-			elog(ERROR, "_mdfd_getrelnfd: cannot open relation %s: %m",
+			elog(ERROR, "could not open relation \"%s\": %m",
 				 RelationGetRelationName(reln));
 		reln->rd_fd = fd;
 	}
@@ -898,7 +918,7 @@ _mdfd_getseg(Relation reln, BlockNumber blkno)
 			v->mdfd_chain = _mdfd_openseg(reln, i, (segno == 1) ? O_CREAT : 0);
 
 			if (v->mdfd_chain == (MdfdVec *) NULL)
-				elog(ERROR, "cannot open segment %u of relation %s (target block %u): %m",
+				elog(ERROR, "could not open segment %u of relation \"%s\" (target block %u): %m",
 					 i, RelationGetRelationName(reln), blkno);
 		}
 		v = v->mdfd_chain;
@@ -951,7 +971,7 @@ _mdfd_blind_getseg(RelFileNode rnode, BlockNumber blkno)
 	/* call fd.c to allow other FDs to be closed if needed */
 	fd = BasicOpenFile(path, O_RDWR | PG_BINARY, 0600);
 	if (fd < 0)
-		elog(LOG, "_mdfd_blind_getseg: couldn't open %s: %m", path);
+		elog(LOG, "could not open \"%s\": %m", path);
 
 	pfree(path);
 
