@@ -2,18 +2,17 @@
  *
  * findbe.c
  *
- * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/utils/init/Attic/findbe.c,v 1.26 2002/02/08 16:30:11 momjian Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/utils/init/findbe.c,v 1.37 2003/08/04 02:40:06 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
 #include "postgres.h"
 
-#include <sys/types.h>
 #include <grp.h>
 #include <pwd.h>
 #include <sys/stat.h>
@@ -45,9 +44,12 @@ static int
 ValidateBinary(char *path)
 {
 	struct stat buf;
+
+#ifndef WIN32
 	uid_t		euid;
 	struct group *gp;
 	struct passwd *pwp;
+#endif
 	int			i;
 	int			is_r = 0;
 	int			is_x = 0;
@@ -61,17 +63,13 @@ ValidateBinary(char *path)
 	 */
 	if (stat(path, &buf) < 0)
 	{
-		if (DebugLvl > 1)
-			fprintf(stderr, "ValidateBinary: can't stat \"%s\"\n",
-					path);
+		elog(DEBUG3, "could not stat \"%s\": %m", path);
 		return -1;
 	}
 
 	if ((buf.st_mode & S_IFMT) != S_IFREG)
 	{
-		if (DebugLvl > 1)
-			fprintf(stderr, "ValidateBinary: \"%s\" is not a regular file\n",
-					path);
+		elog(DEBUG3, "\"%s\" is not a regular file", path);
 		return -1;
 	}
 
@@ -87,14 +85,18 @@ ValidateBinary(char *path)
 	 * Ensure that the file is both executable and readable (required for
 	 * dynamic loading).
 	 */
+#ifdef WIN32
+	is_r = buf.st_mode & S_IRUSR;
+	is_x = buf.st_mode & S_IXUSR;
+	return is_x ? (is_r ? 0 : -2) : -1;
+#else
 	euid = geteuid();
 	if (euid == buf.st_uid)
 	{
 		is_r = buf.st_mode & S_IRUSR;
 		is_x = buf.st_mode & S_IXUSR;
-		if (DebugLvl > 1 && !(is_r && is_x))
-			fprintf(stderr, "ValidateBinary: \"%s\" is not user read/execute\n",
-					path);
+		if (!(is_r && is_x))
+			elog(DEBUG3, "\"%s\" is not user read/execute", path);
 		return is_x ? (is_r ? 0 : -2) : -1;
 	}
 	pwp = getpwuid(euid);
@@ -119,18 +121,17 @@ ValidateBinary(char *path)
 		{
 			is_r = buf.st_mode & S_IRGRP;
 			is_x = buf.st_mode & S_IXGRP;
-			if (DebugLvl > 1 && !(is_r && is_x))
-				fprintf(stderr, "ValidateBinary: \"%s\" is not group read/execute\n",
-						path);
+			if (!(is_r && is_x))
+				elog(DEBUG3, "\"%s\" is not group read/execute", path);
 			return is_x ? (is_r ? 0 : -2) : -1;
 		}
 	}
 	is_r = buf.st_mode & S_IROTH;
 	is_x = buf.st_mode & S_IXOTH;
-	if (DebugLvl > 1 && !(is_r && is_x))
-		fprintf(stderr, "ValidateBinary: \"%s\" is not other read/execute\n",
-				path);
+	if (!(is_r && is_x))
+		elog(DEBUG3, "\"%s\" is not other read/execute", path);
 	return is_x ? (is_r ? 0 : -2) : -1;
+#endif
 }
 
 /*
@@ -165,25 +166,22 @@ FindExec(char *full_path, const char *argv0, const char *binary_name)
 	 * (making sure that a relative path is made absolute before returning
 	 * it).
 	 */
-	if (argv0 && (p = strrchr(argv0, '/')) && *++p)
+	if (argv0 && (p = last_path_separator(argv0)) && *++p)
 	{
-		if (*argv0 == '/' || !getcwd(buf, MAXPGPATH))
+		if (is_absolute_path(argv0) || !getcwd(buf, MAXPGPATH))
 			buf[0] = '\0';
 		else
 			strcat(buf, "/");
 		strcat(buf, argv0);
-		p = strrchr(buf, '/');
+		p = last_path_separator(buf);
 		strcpy(++p, binary_name);
 		if (ValidateBinary(buf) == 0)
 		{
 			strncpy(full_path, buf, MAXPGPATH);
-			if (DebugLvl)
-				fprintf(stderr, "FindExec: found \"%s\" using argv[0]\n",
-						full_path);
+			elog(DEBUG2, "found \"%s\" using argv[0]", full_path);
 			return 0;
 		}
-		fprintf(stderr, "FindExec: invalid binary \"%s\"\n",
-				buf);
+		elog(DEBUG2, "invalid binary \"%s\"", buf);
 		return -1;
 	}
 
@@ -193,8 +191,7 @@ FindExec(char *full_path, const char *argv0, const char *binary_name)
 	 */
 	if ((p = getenv("PATH")) && *p)
 	{
-		if (DebugLvl)
-			fprintf(stderr, "FindExec: searching PATH ...\n");
+		elog(DEBUG2, "searching PATH for executable");
 		path = strdup(p);		/* make a modifiable copy */
 		for (startp = path, endp = strchr(path, ':');
 			 startp && *startp;
@@ -204,7 +201,7 @@ FindExec(char *full_path, const char *argv0, const char *binary_name)
 				continue;
 			if (endp)
 				*endp = '\0';
-			if (*startp == '/' || !getcwd(buf, MAXPGPATH))
+			if (is_absolute_path(startp) || !getcwd(buf, MAXPGPATH))
 				buf[0] = '\0';
 			else
 				strcat(buf, "/");
@@ -215,16 +212,13 @@ FindExec(char *full_path, const char *argv0, const char *binary_name)
 			{
 				case 0: /* found ok */
 					strncpy(full_path, buf, MAXPGPATH);
-					if (DebugLvl)
-						fprintf(stderr, "FindExec: found \"%s\" using PATH\n",
-								full_path);
+					elog(DEBUG2, "found \"%s\" using PATH", full_path);
 					free(path);
 					return 0;
 				case -1:		/* wasn't even a candidate, keep looking */
 					break;
 				case -2:		/* found but disqualified */
-					fprintf(stderr, "FindExec: could not read binary \"%s\"\n",
-							buf);
+					elog(DEBUG2, "could not read binary \"%s\"", buf);
 					free(path);
 					return -1;
 			}
@@ -234,6 +228,6 @@ FindExec(char *full_path, const char *argv0, const char *binary_name)
 		free(path);
 	}
 
-	fprintf(stderr, "FindExec: could not find a %s to execute...\n", binary_name);
+	elog(DEBUG2, "could not find a \"%s\" to execute", binary_name);
 	return -1;
 }
