@@ -15,7 +15,7 @@
  *
  *
  * IDENTIFICATION
- *		$Header: /cvsroot/pgsql/src/bin/pg_dump/pg_backup_archiver.c,v 1.62.2.7 2005/05/17 17:31:15 tgl Exp $
+ *		$Header: /cvsroot/pgsql-server/src/bin/pg_dump/pg_backup_archiver.c,v 1.79.2.2 2004/02/24 03:35:43 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -48,9 +48,11 @@ static ArchiveHandle *_allocAH(const char *FileSpec, const ArchiveFormat fmt,
 		 const int compression, ArchiveMode mode);
 static int	_printTocEntry(ArchiveHandle *AH, TocEntry *te, RestoreOptions *ropt, bool isData);
 
+static void _doSetFixedOutputState(ArchiveHandle *AH);
 static void _doSetSessionAuth(ArchiveHandle *AH, const char *user);
-static void _reconnectAsOwner(ArchiveHandle *AH, const char *dbname, TocEntry *te);
-static void _reconnectAsUser(ArchiveHandle *AH, const char *dbname, const char *user);
+static void _reconnectToDB(ArchiveHandle *AH, const char *dbname, const char *user);
+static void _becomeUser(ArchiveHandle *AH, const char *user);
+static void _becomeOwner(ArchiveHandle *AH, TocEntry *te);
 static void _selectOutputSchema(ArchiveHandle *AH, const char *schemaName);
 
 static teReqs _tocEntryRequired(TocEntry *te, RestoreOptions *ropt);
@@ -119,7 +121,7 @@ CloseArchive(Archive *AHX)
 		res = fclose(AH->OF);
 
 	if (res != 0)
-		die_horribly(AH, modulename, "could not close the output file in CloseArchive\n");
+		die_horribly(AH, modulename, "could not close output archive file\n");
 }
 
 /* Public */
@@ -144,10 +146,6 @@ RestoreArchive(Archive *AHX, RestoreOptions *ropt)
 	 * initially connected to, not the one we will create, which is very
 	 * bad...
 	 */
-
-	if (ropt->create && ropt->noReconnect)
-		die_horribly(AH, modulename, "-C and -R are incompatible options\n");
-
 	if (ropt->create && ropt->dropSchema)
 		die_horribly(AH, modulename, "-C and -c are incompatible options\n");
 
@@ -207,6 +205,11 @@ RestoreArchive(Archive *AHX, RestoreOptions *ropt)
 
 	ahprintf(AH, "--\n-- PostgreSQL database dump\n--\n\n");
 
+	/*
+	 * Establish important parameter values right away.
+	 */
+	_doSetFixedOutputState(AH);
+
 	/*
 	 * Drop the items at the start, in reverse order
 	 */
@@ -221,7 +224,7 @@ RestoreArchive(Archive *AHX, RestoreOptions *ropt)
 				/* We want the schema */
 				ahlog(AH, 1, "dropping %s %s\n", te->desc, te->tag);
 				/* Select owner and schema as necessary */
-				_reconnectAsOwner(AH, NULL, te);
+				_becomeOwner(AH, te);
 				_selectOutputSchema(AH, te->namespace);
 				/* Drop it */
 				ahprintf(AH, "%s", te->dropStmt);
@@ -259,8 +262,8 @@ RestoreArchive(Archive *AHX, RestoreOptions *ropt)
 			/* If we created a DB, connect to it... */
 			if (strcmp(te->desc, "DATABASE") == 0)
 			{
-				ahlog(AH, 1, "connecting to new database %s as user %s\n", te->tag, te->owner);
-				_reconnectAsUser(AH, te->tag, te->owner);
+				ahlog(AH, 1, "connecting to new database \"%s\" as user \"%s\"\n", te->tag, te->owner);
+				_reconnectToDB(AH, te->tag, te->owner);
 			}
 		}
 
@@ -280,11 +283,11 @@ RestoreArchive(Archive *AHX, RestoreOptions *ropt)
 				/*
 				 * If we can output the data, then restore it.
 				 */
-				if (AH->PrintTocDataPtr != NULL && (reqs & REQ_DATA) != 0)
+				if (AH->PrintTocDataPtr !=NULL && (reqs & REQ_DATA) != 0)
 				{
 #ifndef HAVE_LIBZ
 					if (AH->compression != 0)
-						die_horribly(AH, modulename, "unable to restore from compressed archive (not configured for compression support)\n");
+						die_horribly(AH, modulename, "cannot restore from compressed archive (not configured for compression support)\n");
 #endif
 
 					_printTocEntry(AH, te, ropt, true);
@@ -304,20 +307,17 @@ RestoreArchive(Archive *AHX, RestoreOptions *ropt)
 						 * we don't want warnings.
 						 */
 						if (!AH->CustomOutPtr)
-							write_msg(modulename, "WARNING: skipping large object restoration\n");
+							write_msg(modulename, "WARNING: skipping large-object restoration\n");
 					}
 					else
 					{
 						_disableTriggersIfNecessary(AH, te, ropt);
 
-						/*
-						 * Reconnect if necessary (_disableTriggers may
-						 * have reconnected)
-						 */
-						_reconnectAsOwner(AH, NULL, te);
+						/* Select owner and schema as necessary */
+						_becomeOwner(AH, te);
 						_selectOutputSchema(AH, te->namespace);
 
-						ahlog(AH, 1, "restoring data for table %s\n", te->tag);
+						ahlog(AH, 1, "restoring data for table \"%s\"\n", te->tag);
 
 						/*
 						 * If we have a copy statement, use it. As of
@@ -331,7 +331,7 @@ RestoreArchive(Archive *AHX, RestoreOptions *ropt)
 						 * mode with libpq.
 						 */
 						if (te->copyStmt && strlen(te->copyStmt) > 0)
-							ahprintf(AH, "%s", te->copyStmt);
+							ahprintf(AH, te->copyStmt);
 
 						(*AH->PrintTocDataPtr) (AH, te, ropt);
 
@@ -347,7 +347,7 @@ RestoreArchive(Archive *AHX, RestoreOptions *ropt)
 			}
 		}
 		te = te->next;
-	}
+	}		/* end loop over TOC entries */
 
 	/*
 	 * Now use blobs_xref (if used) to fixup any refs for tables that we
@@ -364,18 +364,18 @@ RestoreArchive(Archive *AHX, RestoreOptions *ropt)
 			/* Is it table data? */
 			if (strcmp(te->desc, "TABLE DATA") == 0)
 			{
-				ahlog(AH, 2, "checking whether we loaded %s\n", te->tag);
+				ahlog(AH, 2, "checking whether we loaded \"%s\"\n", te->tag);
 
 				reqs = _tocEntryRequired(te, ropt);
 
 				if ((reqs & REQ_DATA) != 0)		/* We loaded the data */
 				{
-					ahlog(AH, 1, "fixing up large object cross-reference for %s\n", te->tag);
+					ahlog(AH, 1, "fixing up large-object cross-reference for \"%s\"\n", te->tag);
 					FixupBlobRefs(AH, te);
 				}
 			}
 			else
-				ahlog(AH, 2, "ignoring large object cross-references for %s %s\n", te->desc, te->tag);
+				ahlog(AH, 2, "ignoring large-object cross-references for %s %s\n", te->desc, te->tag);
 
 			te = te->next;
 		}
@@ -387,7 +387,7 @@ RestoreArchive(Archive *AHX, RestoreOptions *ropt)
 	/*
 	 * Clean up & we're done.
 	 */
-	if (ropt->filename)
+	if (ropt->filename || ropt->compression)
 		ResetOutput(AH, sav);
 
 	if (ropt->useDB)
@@ -439,9 +439,6 @@ _canRestoreBlobs(ArchiveHandle *AH)
 static void
 _disableTriggersIfNecessary(ArchiveHandle *AH, TocEntry *te, RestoreOptions *ropt)
 {
-	char	   *oldUser;
-	char	   *oldSchema;
-
 	/* This hack is only needed in a data-only restore */
 	if (!ropt->dataOnly || !ropt->disable_triggers)
 		return;
@@ -450,23 +447,12 @@ _disableTriggersIfNecessary(ArchiveHandle *AH, TocEntry *te, RestoreOptions *rop
 	if (te && strcmp(te->desc, "BLOBS") == 0)
 		return;
 
-	oldUser = strdup(AH->currUser);
-	oldSchema = strdup(AH->currSchema);
-
 	/*
 	 * Become superuser if possible, since they are the only ones who can
-	 * update pg_class.  If -S was not given, but we are allowed to use
-	 * SET SESSION AUTHORIZATION, assume the initial user identity is a
-	 * superuser.  Otherwise we just have to bull ahead anyway.
+	 * update pg_class.  If -S was not given, assume the initial user identity
+	 * is a superuser.
 	 */
-	if (ropt->superuser)
-	{
-		_reconnectAsUser(AH, NULL, ropt->superuser);
-		/* be careful to preserve schema setting */
-		_selectOutputSchema(AH, oldSchema);
-	}
-	else if (AH->ropt->use_setsessauth)
-		_doSetSessionAuth(AH, NULL);
+	_becomeUser(AH, ropt->superuser);
 
 	ahlog(AH, 1, "disabling triggers\n");
 
@@ -487,28 +473,11 @@ _disableTriggersIfNecessary(ArchiveHandle *AH, TocEntry *te, RestoreOptions *rop
 	else
 		ahprintf(AH, "UPDATE pg_catalog.pg_class SET reltriggers = 0 FROM pg_catalog.pg_namespace "
 				 "WHERE relnamespace = pg_namespace.oid AND nspname !~ '^pg_';\n\n");
-
-	/*
-	 * Restore original user and schema state.
-	 */
-	if (ropt->superuser)
-	{
-		_reconnectAsUser(AH, NULL, oldUser);
-		/* be careful to preserve schema setting */
-		_selectOutputSchema(AH, oldSchema);
-	}
-	else if (AH->ropt->use_setsessauth)
-		_doSetSessionAuth(AH, oldUser);
-	free(oldUser);
-	free(oldSchema);
 }
 
 static void
 _enableTriggersIfNecessary(ArchiveHandle *AH, TocEntry *te, RestoreOptions *ropt)
 {
-	char	   *oldUser;
-	char	   *oldSchema;
-
 	/* This hack is only needed in a data-only restore */
 	if (!ropt->dataOnly || !ropt->disable_triggers)
 		return;
@@ -517,23 +486,12 @@ _enableTriggersIfNecessary(ArchiveHandle *AH, TocEntry *te, RestoreOptions *ropt
 	if (te && strcmp(te->desc, "BLOBS") == 0)
 		return;
 
-	oldUser = strdup(AH->currUser);
-	oldSchema = strdup(AH->currSchema);
-
 	/*
 	 * Become superuser if possible, since they are the only ones who can
-	 * update pg_class.  If -S was not given, but we are allowed to use
-	 * SET SESSION AUTHORIZATION, assume the initial user identity is a
-	 * superuser.  Otherwise we just have to bull ahead anyway.
+	 * update pg_class.  If -S was not given, assume the initial user identity
+	 * is a superuser.
 	 */
-	if (ropt->superuser)
-	{
-		_reconnectAsUser(AH, NULL, ropt->superuser);
-		/* be careful to preserve schema setting */
-		_selectOutputSchema(AH, oldSchema);
-	}
-	else if (AH->ropt->use_setsessauth)
-		_doSetSessionAuth(AH, NULL);
+	_becomeUser(AH, ropt->superuser);
 
 	ahlog(AH, 1, "enabling triggers\n");
 
@@ -557,20 +515,6 @@ _enableTriggersIfNecessary(ArchiveHandle *AH, TocEntry *te, RestoreOptions *ropt
 				 "(SELECT pg_catalog.count(*) FROM pg_catalog.pg_trigger where pg_class.oid = tgrelid) "
 				 "FROM pg_catalog.pg_namespace "
 				 "WHERE relnamespace = pg_namespace.oid AND nspname !~ '^pg_';\n\n");
-
-	/*
-	 * Restore original user and schema state.
-	 */
-	if (ropt->superuser)
-	{
-		_reconnectAsUser(AH, NULL, oldUser);
-		/* be careful to preserve schema setting */
-		_selectOutputSchema(AH, oldSchema);
-	}
-	else if (AH->ropt->use_setsessauth)
-		_doSetSessionAuth(AH, oldUser);
-	free(oldUser);
-	free(oldSchema);
 }
 
 /*
@@ -584,7 +528,7 @@ WriteData(Archive *AHX, const void *data, size_t dLen)
 	ArchiveHandle *AH = (ArchiveHandle *) AHX;
 
 	if (!AH->currToc)
-		die_horribly(AH, modulename, "WriteData cannot be called outside the context of a DataDumper routine\n");
+		die_horribly(AH, modulename, "internal error -- WriteData cannot be called outside the context of a DataDumper routine\n");
 
 	return (*AH->WriteDataPtr) (AH, data, dLen);
 }
@@ -658,7 +602,7 @@ PrintTOCSummary(Archive *AHX, RestoreOptions *ropt)
 	char	   *fmtName;
 
 	if (ropt->filename)
-		sav = SetOutput(AH, ropt->filename, ropt->compression);
+		sav = SetOutput(AH, ropt->filename, 0 /* no compression */);
 
 	ahprintf(AH, ";\n; Archive created at %s", ctime(&AH->createDate));
 	ahprintf(AH, ";     dbname: %s\n;     TOC Entries: %d\n;     Compression: %d\n",
@@ -708,7 +652,7 @@ StartBlob(Archive *AHX, Oid oid)
 	ArchiveHandle *AH = (ArchiveHandle *) AHX;
 
 	if (!AH->StartBlobPtr)
-		die_horribly(AH, modulename, "large object output not supported in chosen format\n");
+		die_horribly(AH, modulename, "large-object output not supported in chosen format\n");
 
 	(*AH->StartBlobPtr) (AH, AH->currToc, oid);
 
@@ -748,7 +692,7 @@ EndRestoreBlobs(ArchiveHandle *AH)
 {
 	if (AH->txActive)
 	{
-		ahlog(AH, 2, "committing large object transactions\n");
+		ahlog(AH, 2, "committing large-object transactions\n");
 		CommitTransaction(AH);
 	}
 
@@ -786,7 +730,7 @@ StartRestoreBlob(ArchiveHandle *AH, Oid oid)
 	 */
 	if (!AH->txActive)
 	{
-		ahlog(AH, 2, "starting large object transactions\n");
+		ahlog(AH, 2, "starting large-object transactions\n");
 		StartTransaction(AH);
 	}
 	if (!AH->blobTxActive)
@@ -796,7 +740,7 @@ StartRestoreBlob(ArchiveHandle *AH, Oid oid)
 	if (loOid == 0)
 		die_horribly(AH, modulename, "could not create large object\n");
 
-	ahlog(AH, 2, "restoring large object with oid %u as %u\n", oid, loOid);
+	ahlog(AH, 2, "restoring large object with OID %u as %u\n", oid, loOid);
 
 	InsertBlobXref(AH, oid, loOid);
 
@@ -817,7 +761,7 @@ EndRestoreBlob(ArchiveHandle *AH, Oid oid)
 
 		res = lo_write(AH->connection, AH->loFd, (void *) AH->lo_buf, AH->lo_buf_used);
 
-		ahlog(AH, 5, "wrote remaining %lu bytes of large object data (result = %lu)\n",
+		ahlog(AH, 5, "wrote remaining %lu bytes of large-object data (result = %lu)\n",
 			  (unsigned long) AH->lo_buf_used, (unsigned long) res);
 		if (res != AH->lo_buf_used)
 			die_horribly(AH, modulename, "could not write to large object (result: %lu, expected: %lu)\n",
@@ -833,7 +777,7 @@ EndRestoreBlob(ArchiveHandle *AH, Oid oid)
 	 */
 	if (((AH->blobCount / BLOB_BATCH_SIZE) * BLOB_BATCH_SIZE) == AH->blobCount)
 	{
-		ahlog(AH, 2, "committing large object transactions\n");
+		ahlog(AH, 2, "committing large-object transactions\n");
 		CommitTransaction(AH);
 		CommitTransactionXref(AH);
 	}
@@ -845,11 +789,12 @@ EndRestoreBlob(ArchiveHandle *AH, Oid oid)
 
 /*
  * Move TOC entries of the specified type to the START of the TOC.
+ *
+ * This is public, but if you use it anywhere but SortTocByObjectType,
+ * you are risking breaking things.
  */
-
-/* Public */
 void
-MoveToStart(Archive *AHX, char *oType)
+MoveToStart(Archive *AHX, const char *oType)
 {
 	ArchiveHandle *AH = (ArchiveHandle *) AHX;
 	TocEntry   *te = AH->toc->next;
@@ -874,10 +819,12 @@ MoveToStart(Archive *AHX, char *oType)
 
 /*
  * Move TOC entries of the specified type to the end of the TOC.
+ *
+ * This is public, but if you use it anywhere but SortTocByObjectType,
+ * you are risking breaking things.
  */
-/* Public */
 void
-MoveToEnd(Archive *AHX, char *oType)
+MoveToEnd(Archive *AHX, const char *oType)
 {
 	ArchiveHandle *AH = (ArchiveHandle *) AHX;
 	TocEntry   *te = AH->toc->next;
@@ -899,6 +846,44 @@ MoveToEnd(Archive *AHX, char *oType)
 	}
 }
 
+/*
+ * Sort TOC by object type (items of same type keep same relative order)
+ *
+ * This is factored out to ensure that pg_dump and pg_restore stay in sync
+ * about the standard ordering.
+ */
+void
+SortTocByObjectType(Archive *AH)
+{
+	/*
+	 * Procedural languages have to be declared just after database and
+	 * schema creation, before they are used.
+	 */
+	MoveToStart(AH, "ACL LANGUAGE");
+	MoveToStart(AH, "PROCEDURAL LANGUAGE");
+	MoveToStart(AH, "FUNC PROCEDURAL LANGUAGE");
+	MoveToStart(AH, "SCHEMA");
+	MoveToStart(AH, "<Init>");
+	/* Database entries *must* be at front (see also pg_restore.c) */
+	MoveToStart(AH, "DATABASE");
+
+	MoveToEnd(AH, "TABLE DATA");
+	MoveToEnd(AH, "BLOBS");
+	MoveToEnd(AH, "INDEX");
+	MoveToEnd(AH, "CONSTRAINT");
+	MoveToEnd(AH, "FK CONSTRAINT");
+	MoveToEnd(AH, "TRIGGER");
+	MoveToEnd(AH, "RULE");
+	MoveToEnd(AH, "SEQUENCE SET");
+
+	/*
+	 * Moving all comments to end is annoying, but must do it for comments
+	 * on stuff we just moved, and we don't seem to have quite enough
+	 * dependency structure to get it really right...
+	 */
+	MoveToEnd(AH, "COMMENT");
+}
+
 /*
  * Sort TOC by OID
  */
@@ -959,7 +944,7 @@ SortTocFromFile(Archive *AHX, RestoreOptions *ropt)
 	if (!fh)
 		die_horribly(AH, modulename, "could not open TOC file\n");
 
-	while (fgets(buf, sizeof(buf), fh) != NULL)
+	while (fgets(buf, 1024, fh) != NULL)
 	{
 		/* Find a comment */
 		cmnt = strchr(buf, ';');
@@ -985,15 +970,12 @@ SortTocFromFile(Archive *AHX, RestoreOptions *ropt)
 		/* Find TOC entry */
 		te = _getTocEntry(AH, id);
 		if (!te)
-			die_horribly(AH, modulename, "could not find entry for id %d\n", id);
+			die_horribly(AH, modulename, "could not find entry for ID %d\n", id);
 
-		if (!ropt->idWanted[id - 1])
-		{
-			ropt->idWanted[id - 1] = 1;
+		ropt->idWanted[id - 1] = 1;
 
-			_moveAfter(AH, tePrev, te);
-			tePrev = te;
-		}
+		_moveAfter(AH, tePrev, te);
+		tePrev = te;
 	}
 
 	if (fclose(fh) != 0)
@@ -1063,23 +1045,19 @@ OutputContext
 SetOutput(ArchiveHandle *AH, char *filename, int compression)
 {
 	OutputContext sav;
-
-#ifdef HAVE_LIBZ
-	char		fmode[10];
-#endif
-	int			fn = 0;
+	int			fn;
 
 	/* Replace the AH output file handle */
 	sav.OF = AH->OF;
 	sav.gzOut = AH->gzOut;
 
 	if (filename)
-		fn = 0;
+		fn = -1;
 	else if (AH->FH)
 		fn = fileno(AH->FH);
 	else if (AH->fSpec)
 	{
-		fn = 0;
+		fn = -1;
 		filename = AH->fSpec;
 	}
 	else
@@ -1089,27 +1067,25 @@ SetOutput(ArchiveHandle *AH, char *filename, int compression)
 #ifdef HAVE_LIBZ
 	if (compression != 0)
 	{
+		char		fmode[10];
+
+		/* Don't use PG_BINARY_x since this is zlib */
 		sprintf(fmode, "wb%d", compression);
-		if (fn)
-		{
-			AH->OF = gzdopen(dup(fn), fmode);	/* Don't use PG_BINARY_x
-												 * since this is zlib */
-		}
+		if (fn >= 0)
+			AH->OF = gzdopen(dup(fn), fmode);
 		else
 			AH->OF = gzopen(filename, fmode);
 		AH->gzOut = 1;
 	}
 	else
-	{							/* Use fopen */
 #endif
-		if (fn)
+	{							/* Use fopen */
+		if (fn >= 0)
 			AH->OF = fdopen(dup(fn), PG_BINARY_W);
 		else
 			AH->OF = fopen(filename, PG_BINARY_W);
 		AH->gzOut = 0;
-#ifdef HAVE_LIBZ
 	}
-#endif
 
 	if (!AH->OF)
 		die_horribly(AH, modulename, "could not open output file: %s\n", strerror(errno));
@@ -1128,7 +1104,8 @@ ResetOutput(ArchiveHandle *AH, OutputContext sav)
 		res = fclose(AH->OF);
 
 	if (res != 0)
-		die_horribly(AH, modulename, "could not close output file: %s\n", strerror(errno));
+		die_horribly(AH, modulename, "could not close output file: %s\n",
+					 strerror(errno));
 
 	AH->gzOut = sav.gzOut;
 	AH->OF = sav.OF;
@@ -1295,7 +1272,8 @@ _die_horribly(ArchiveHandle *AH, const char *modulename, const char *fmt, va_lis
 {
 	_write_msg(modulename, fmt, ap);
 
-	if (AH) {
+	if (AH)
+	{
 		if (AH->public.verbose)
 			write_msg(NULL, "*** aborted because of error\n");
 		if (AH->connection)
@@ -1386,7 +1364,7 @@ TocIDRequired(ArchiveHandle *AH, int id, RestoreOptions *ropt)
 size_t
 WriteOffset(ArchiveHandle *AH, off_t o, int wasSet)
 {
-	int 			off;
+	int			off;
 
 	/* Save the flag */
 	(*AH->WriteBytePtr) (AH, wasSet);
@@ -1394,7 +1372,7 @@ WriteOffset(ArchiveHandle *AH, off_t o, int wasSet)
 	/* Write out off_t smallest byte first, prevents endian mismatch */
 	for (off = 0; off < sizeof(off_t); off++)
 	{
-	    (*AH->WriteBytePtr) (AH, o & 0xFF);
+		(*AH->WriteBytePtr) (AH, o & 0xFF);
 		o >>= 8;
 	}
 	return sizeof(off_t) + 1;
@@ -1403,9 +1381,9 @@ WriteOffset(ArchiveHandle *AH, off_t o, int wasSet)
 int
 ReadOffset(ArchiveHandle *AH, off_t *o)
 {
-	int				i;
-	int 			off;
-	int				offsetFlg;
+	int			i;
+	int			off;
+	int			offsetFlg;
 
 	/* Initialize to zero */
 	*o = 0;
@@ -1417,21 +1395,21 @@ ReadOffset(ArchiveHandle *AH, off_t *o)
 		i = ReadInt(AH);
 		/* -1 means not set */
 		if (i < 0)
-		    return K_OFFSET_POS_NOT_SET;
+			return K_OFFSET_POS_NOT_SET;
 		else if (i == 0)
-		    return K_OFFSET_NO_DATA;
+			return K_OFFSET_NO_DATA;
 
 		/* Cast to off_t because it was written as an int. */
-		*o = (off_t)i;
+		*o = (off_t) i;
 		return K_OFFSET_POS_SET;
 	}
 
 	/*
-	 * Read the flag indicating the state of the data pointer.
-	 * Check if valid and die if not.
+	 * Read the flag indicating the state of the data pointer. Check if
+	 * valid and die if not.
 	 *
-	 * This used to be handled by a negative or zero pointer,
-	 * now we use an extra byte specifically for the state.
+	 * This used to be handled by a negative or zero pointer, now we use an
+	 * extra byte specifically for the state.
 	 */
 	offsetFlg = (*AH->ReadBytePtr) (AH) & 0xFF;
 
@@ -1441,10 +1419,10 @@ ReadOffset(ArchiveHandle *AH, off_t *o)
 		case K_OFFSET_NO_DATA:
 		case K_OFFSET_POS_SET:
 
-				break;
+			break;
 
 		default:
-				die_horribly(AH, modulename, "Unexpected data offset flag %d\n", offsetFlg);
+			die_horribly(AH, modulename, "Unexpected data offset flag %d\n", offsetFlg);
 	}
 
 	/*
@@ -1457,7 +1435,7 @@ ReadOffset(ArchiveHandle *AH, off_t *o)
 		else
 		{
 			if ((*AH->ReadBytePtr) (AH) != 0)
-	    	    die_horribly(AH, modulename, "file offset in dump file is too large\n");
+				die_horribly(AH, modulename, "file offset in dump file is too large\n");
 		}
 	}
 
@@ -1663,7 +1641,7 @@ _discoverArchiveFormat(ArchiveHandle *AH)
 	if (fseeko(fh, 0, SEEK_SET) != 0)
 	{
 		/*
-		 * NOTE: Formats that use the looahead buffer can unset this in
+		 * NOTE: Formats that use the lookahead buffer can unset this in
 		 * their Init routine.
 		 */
 		AH->readHeader = 1;
@@ -1778,7 +1756,7 @@ _allocAH(const char *FileSpec, const ArchiveFormat fmt,
 			break;
 
 		default:
-			die_horribly(AH, modulename, "unrecognized file format '%d'\n", fmt);
+			die_horribly(AH, modulename, "unrecognized file format \"%d\"\n", fmt);
 	}
 
 	return AH;
@@ -1891,7 +1869,7 @@ ReadToc(ArchiveHandle *AH)
 
 		/* Sanity check */
 		if (te->id <= 0 || te->id > AH->tocCount)
-			die_horribly(AH, modulename, "entry id %d out of range - perhaps a corrupt TOC\n", te->id);
+			die_horribly(AH, modulename, "entry ID %d out of range -- perhaps a corrupt TOC\n", te->id);
 
 		te->hadDumper = ReadInt(AH);
 		te->oid = ReadStr(AH);
@@ -1948,7 +1926,7 @@ ReadToc(ArchiveHandle *AH)
 		if (AH->ReadExtraTocPtr)
 			(*AH->ReadExtraTocPtr) (AH, te);
 
-		ahlog(AH, 3, "read TOC entry %d (id %d) for %s %s\n", i, te->id, te->desc, te->tag);
+		ahlog(AH, 3, "read TOC entry %d (ID %d) for %s %s\n", i, te->id, te->desc, te->tag);
 
 		te->prev = AH->toc->prev;
 		AH->toc->prev->next = te;
@@ -1962,6 +1940,10 @@ _tocEntryRequired(TocEntry *te, RestoreOptions *ropt)
 {
 	teReqs		res = 3;		/* Schema = 1, Data = 2, Both = 3 */
 
+	/* ENCODING objects are dumped specially, so always reject here */
+	if (strcmp(te->desc, "ENCODING") == 0)
+		return 0;
+
 	/* If it's an ACL, maybe ignore it */
 	if (ropt->aclsSkip && strcmp(te->desc, "ACL") == 0)
 		return 0;
@@ -2035,7 +2017,7 @@ _tocEntryRequired(TocEntry *te, RestoreOptions *ropt)
 	if (ropt->dataOnly)
 		res = res & REQ_DATA;
 
-	/* Mask it if we don't have a schema contribition */
+	/* Mask it if we don't have a schema contribution */
 	if (!te->defn || strlen(te->defn) == 0)
 		res = res & ~REQ_SCHEMA;
 
@@ -2046,10 +2028,37 @@ _tocEntryRequired(TocEntry *te, RestoreOptions *ropt)
 	return res;
 }
 
+/*
+ * Issue SET commands for parameters that we want to have set the same way
+ * at all times during execution of a restore script.
+ */
+static void
+_doSetFixedOutputState(ArchiveHandle *AH)
+{
+	TocEntry   *te;
+
+	/* If we have an encoding setting, emit that */
+	te = AH->toc->next;
+	while (te != AH->toc)
+	{
+		if (strcmp(te->desc, "ENCODING") == 0)
+		{
+			ahprintf(AH, "%s", te->defn);
+			break;
+		}
+		te = te->next;
+	}
+
+	/* Make sure function checking is disabled */
+	ahprintf(AH, "SET check_function_bodies = false;\n");
+
+	ahprintf(AH, "\n");
+}
+
 /*
  * Issue a SET SESSION AUTHORIZATION command.  Caller is responsible
- * for updating state if appropriate.  If user is NULL, the
- * specification DEFAULT will be used.
+ * for updating state if appropriate.  If user is NULL or an empty string,
+ * the specification DEFAULT will be used.
  */
 static void
 _doSetSessionAuth(ArchiveHandle *AH, const char *user)
@@ -2057,11 +2066,11 @@ _doSetSessionAuth(ArchiveHandle *AH, const char *user)
 	PQExpBuffer cmd = createPQExpBuffer();
 
 	appendPQExpBuffer(cmd, "SET SESSION AUTHORIZATION ");
-	if (user)
 
-		/*
-		 * SQL requires a string literal here.	Might as well be correct.
-		 */
+	/*
+	 * SQL requires a string literal here.	Might as well be correct.
+	 */
+	if (user && *user)
 		appendStringLiteral(cmd, user, false);
 	else
 		appendPQExpBuffer(cmd, "DEFAULT");
@@ -2074,7 +2083,7 @@ _doSetSessionAuth(ArchiveHandle *AH, const char *user)
 		res = PQexec(AH->connection, cmd->data);
 
 		if (!res || PQresultStatus(res) != PGRES_COMMAND_OK)
-			die_horribly(AH, modulename, "could not set session user to %s: %s",
+			die_horribly(AH, modulename, "could not set session user to \"%s\": %s",
 						 user, PQerrorMessage(AH->connection));
 
 		PQclear(res);
@@ -2087,34 +2096,17 @@ _doSetSessionAuth(ArchiveHandle *AH, const char *user)
 
 
 /*
- * Issue the commands to connect to the database as the specified user
- * to the specified database.  The database name may be NULL, then the
- * current database is kept.  If reconnects were disallowed by the
- * user, this won't do anything.
+ * Issue the commands to connect to the specified database
+ * as the specified user.
  *
  * If we're currently restoring right into a database, this will
  * actually establish a connection. Otherwise it puts a \connect into
  * the script output.
  */
 static void
-_reconnectAsUser(ArchiveHandle *AH, const char *dbname, const char *user)
+_reconnectToDB(ArchiveHandle *AH, const char *dbname, const char *user)
 {
-	if (!user || strlen(user) == 0
-		|| (strcmp(AH->currUser, user) == 0 && !dbname))
-		return;					/* no need to do anything */
-
-	/*
-	 * Use SET SESSION AUTHORIZATION if allowed and no database change
-	 * needed
-	 */
-	if (!dbname && AH->ropt->use_setsessauth)
-		_doSetSessionAuth(AH, user);
-	else if (AH->ropt && AH->ropt->noReconnect)
-	{
-		/* When -R was given, don't do anything. */
-		return;
-	}
-	else if (RestoringToDB(AH))
+	if (RestoringToDB(AH))
 		ReconnectToServer(AH, dbname, user);
 	else
 	{
@@ -2125,14 +2117,9 @@ _reconnectAsUser(ArchiveHandle *AH, const char *dbname, const char *user)
 		appendPQExpBuffer(qry, " %s\n\n",
 						  fmtId(user));
 
-		ahprintf(AH, "%s", qry->data);
+		ahprintf(AH, qry->data);
 
 		destroyPQExpBuffer(qry);
-
-		/* don't assume we still know the output schema */
-		if (AH->currSchema)
-			free(AH->currSchema);
-		AH->currSchema = strdup("");
 	}
 
 	/*
@@ -2143,21 +2130,53 @@ _reconnectAsUser(ArchiveHandle *AH, const char *dbname, const char *user)
 		free(AH->currUser);
 
 	AH->currUser = strdup(user);
+
+	/* don't assume we still know the output schema */
+	if (AH->currSchema)
+		free(AH->currSchema);
+	AH->currSchema = strdup("");
+
+	/* re-establish fixed state */
+	_doSetFixedOutputState(AH);
+}
+
+/*
+ * Become the specified user, and update state to avoid redundant commands
+ *
+ * NULL or empty argument is taken to mean restoring the session default
+ */
+static void
+_becomeUser(ArchiveHandle *AH, const char *user)
+{
+	if (!user)
+		user = "";				/* avoid null pointers */
+
+	if (AH->currUser && strcmp(AH->currUser, user) == 0)
+		return;					/* no need to do anything */
+
+	_doSetSessionAuth(AH, user);
+
+	/*
+	 * NOTE: currUser keeps track of what the imaginary session user in
+	 * our script is
+	 */
+	if (AH->currUser)
+		free(AH->currUser);
+
+	AH->currUser = strdup(user);
 }
 
-
 /*
- * Issues the commands to connect to the database (or the current one,
- * if NULL) as the owner of the the given TOC entry object.  If
+ * Become the owner of the the given TOC entry object.  If
  * changes in ownership are not allowed, this doesn't do anything.
  */
 static void
-_reconnectAsOwner(ArchiveHandle *AH, const char *dbname, TocEntry *te)
+_becomeOwner(ArchiveHandle *AH, TocEntry *te)
 {
 	if (AH->ropt && AH->ropt->noOwner)
 		return;
 
-	_reconnectAsUser(AH, dbname, te->owner);
+	_becomeUser(AH, te->owner);
 }
 
 
@@ -2188,7 +2207,7 @@ _selectOutputSchema(ArchiveHandle *AH, const char *schemaName)
 		res = PQexec(AH->connection, qry->data);
 
 		if (!res || PQresultStatus(res) != PGRES_COMMAND_OK)
-			die_horribly(AH, modulename, "could not set search_path to %s: %s",
+			die_horribly(AH, modulename, "could not set search_path to \"%s\": %s",
 						 schemaName, PQerrorMessage(AH->connection));
 
 		PQclear(res);
@@ -2211,7 +2230,7 @@ _printTocEntry(ArchiveHandle *AH, TocEntry *te, RestoreOptions *ropt, bool isDat
 	char	   *pfx;
 
 	/* Select owner and schema as necessary */
-	_reconnectAsOwner(AH, NULL, te);
+	_becomeOwner(AH, te);
 	_selectOutputSchema(AH, te->namespace);
 
 	if (isData)
@@ -2227,8 +2246,21 @@ _printTocEntry(ArchiveHandle *AH, TocEntry *te, RestoreOptions *ropt, bool isDat
 		(*AH->PrintExtraTocPtr) (AH, te);
 	ahprintf(AH, "--\n\n");
 
-	if (strlen(te->defn) > 0)
-		ahprintf(AH, "%s\n\n", te->defn);
+	/*
+	 * Really crude hack for suppressing AUTHORIZATION clause of CREATE SCHEMA
+	 * when --no-owner mode is selected.  This is ugly, but I see no other
+	 * good way ...
+	 */
+	if (AH->ropt && AH->ropt->noOwner && strcmp(te->desc, "SCHEMA") == 0)
+	{
+		ahprintf(AH, "CREATE SCHEMA %s;\n\n\n", te->tag);
+	}
+	else
+	{
+		/* normal case */
+		if (strlen(te->defn) > 0)
+			ahprintf(AH, "%s\n\n", te->defn);
+	}
 
 	return 1;
 }
@@ -2249,7 +2281,7 @@ WriteHead(ArchiveHandle *AH)
 #ifndef HAVE_LIBZ
 	if (AH->compression != 0)
 		write_msg(modulename, "WARNING: requested compression not available in this "
-				  "installation - archive will be uncompressed\n");
+				  "installation -- archive will be uncompressed\n");
 
 	AH->compression = 0;
 #endif
@@ -2307,9 +2339,9 @@ ReadHead(ArchiveHandle *AH)
 			write_msg(modulename, "WARNING: archive was made on a machine with larger integers, some operations may fail\n");
 
 		if (AH->version >= K_VERS_1_7)
-		    AH->offSize = (*AH->ReadBytePtr) (AH);
+			AH->offSize = (*AH->ReadBytePtr) (AH);
 		else
-		    AH->offSize = AH->intSize;
+			AH->offSize = AH->intSize;
 
 		fmt = (*AH->ReadBytePtr) (AH);
 
@@ -2330,7 +2362,7 @@ ReadHead(ArchiveHandle *AH)
 
 #ifndef HAVE_LIBZ
 	if (AH->compression != 0)
-		write_msg(modulename, "WARNING: archive is compressed, but this installation does not support compression - no data will be available\n");
+		write_msg(modulename, "WARNING: archive is compressed, but this installation does not support compression -- no data will be available\n");
 #endif
 
 	if (AH->version >= K_VERS_1_4)
@@ -2366,10 +2398,11 @@ checkSeek(FILE *fp)
 	if (fseeko(fp, 0, SEEK_CUR) != 0)
 		return false;
 	else if (sizeof(off_t) > sizeof(long))
-	/*
-	 *	At this point, off_t is too large for long, so we return
-	 *	based on whether an off_t version of fseek is available.
-	 */
+
+		/*
+		 * At this point, off_t is too large for long, so we return based
+		 * on whether an off_t version of fseek is available.
+		 */
 #ifdef HAVE_FSEEKO
 		return true;
 #else
