@@ -8,7 +8,7 @@
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/interfaces/libpq/fe-exec.c,v 1.153.2.1 2003/11/30 20:53:43 joe Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/interfaces/libpq/fe-exec.c,v 1.153.2.3 2004/03/05 01:54:13 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -609,12 +609,28 @@ pqSaveParameterStatus(PGconn *conn, const char *name, const char *value)
 
 	/*
 	 * Special hacks: remember client_encoding as a numeric value, and
-	 * remember at least the first few bytes of server version.
+	 * convert server version to a numeric form as well.
 	 */
 	if (strcmp(name, "client_encoding") == 0)
 		conn->client_encoding = pg_char_to_encoding(value);
-	if (strcmp(name, "server_version") == 0)
-		StrNCpy(conn->sversion, value, sizeof(conn->sversion));
+	else if (strcmp(name, "server_version") == 0)
+	{
+		int			cnt;
+		int			vmaj,
+					vmin,
+					vrev;
+
+		cnt = sscanf(value, "%d.%d.%d", &vmaj, &vmin, &vrev);
+
+		if (cnt < 2)
+			conn->sversion = 0;			/* unknown */
+		else
+		{
+			if (cnt == 2)
+				vrev = 0;
+			conn->sversion = (100 * vmaj + vmin) * 100 + vrev;
+		}
+	}
 }
 
 
@@ -1234,6 +1250,9 @@ PQexecStart(PGconn *conn)
 				return false;
 			}
 		}
+		/* check for loss of connection, too */
+		if (conn->status == CONNECTION_BAD)
+			return false;
 	}
 
 	/* OK to send a command */
@@ -1256,6 +1275,8 @@ PQexecFinish(PGconn *conn)
 	 *
 	 * We have to stop if we see copy in/out, however. We will resume parsing
 	 * after application performs the data transfer.
+	 *
+	 * Also stop if the connection is lost (else we'll loop infinitely).
 	 */
 	lastResult = NULL;
 	while ((result = PQgetResult(conn)) != NULL)
@@ -1281,7 +1302,8 @@ PQexecFinish(PGconn *conn)
 		}
 		lastResult = result;
 		if (result->resultStatus == PGRES_COPY_IN ||
-			result->resultStatus == PGRES_COPY_OUT)
+			result->resultStatus == PGRES_COPY_OUT ||
+			conn->status == CONNECTION_BAD)
 			break;
 	}
 
